@@ -1,7 +1,10 @@
+from __future__ import absolute_import
+
 import os
 import re
 import hashlib
 import email.utils
+import six
 
 import colander
 import yaml
@@ -9,11 +12,14 @@ import yaml
 from stat import ST_MODE
 from stat import S_IXUSR
 
-from linter import Linter
-from launchpadlib.launchpad import Launchpad
+import dict2colander
 
-KNOWN_METADATA_KEYS = [
+from charmtools.linter import Linter
+from charmtools.utils import validate_display_name
+
+KNOWN_METADATA_KEYS = (
     'name',
+    'display-name',
     'summary',
     'maintainer',
     'maintainers',
@@ -29,16 +35,26 @@ KNOWN_METADATA_KEYS = [
     'series',
     'storage',
     'extra-bindings',
-]
+    'payloads',
+    'terms',
+    'resources',
+    'devices',
+    'deployment',
+)
 
-KNOWN_RELATION_KEYS = ['interface', 'scope', 'limit', 'optional']
+REQUIRED_METADATA_KEYS = (
+    'name',
+    'summary',
+)
 
-KNOWN_SCOPES = ['global', 'container']
+KNOWN_RELATION_KEYS = ('interface', 'scope', 'limit', 'optional')
+
+KNOWN_SCOPES = ('global', 'container')
 
 TEMPLATE_PATH = os.path.abspath(os.path.dirname(__file__))
 
 TEMPLATE_README = os.path.join(
-    TEMPLATE_PATH, 'templates', 'bash', 'files', 'README.ex')
+    TEMPLATE_PATH, 'templates', 'bash', 'files', 'README.md')
 
 TEMPLATE_ICON = os.path.join(
     TEMPLATE_PATH, 'templates', 'bash', 'files', 'icon.svg')
@@ -46,13 +62,13 @@ TEMPLATE_ICON = os.path.join(
 KNOWN_OPTION_KEYS = set(('description', 'type', 'default'))
 
 KNOWN_OPTION_TYPES = {
-    'string': basestring,
+    'string': six.string_types[0],
     'int': int,
     'float': float,
     'boolean': bool,
 }
 
-ALLOW_NONE_DEFAULT = (basestring, int, float)
+ALLOW_NONE_DEFAULT = (six.string_types[0], int, float, bool)
 
 
 class RelationError(Exception):
@@ -91,7 +107,7 @@ class CharmLinter(Linter):
                 for line in hook_file:
                     count += 1
                     hook_warnings = [
-                        {'re': re.compile("http://169\.254\.169\.254/"),
+                        {'re': re.compile(r"http://169\.254\.169\.254/"),
                          'msg': "hook accesses EC2 metadata service directly"}]
                     for warning in hook_warnings:
                         if warning['re'].search(line):
@@ -158,7 +174,7 @@ class CharmLinter(Linter):
             self.info('File config.yaml not found.')
             return
         try:
-            with open(config_path) as config_file:
+            with open(config_path, "rb") as config_file:
                 config = yaml.safe_load(config_file.read())
         except Exception as error:
             self.err('Cannot parse config.yaml: %s' % error)
@@ -181,7 +197,7 @@ class CharmLinter(Linter):
             return
 
         for option_name, option_value in options.items():
-            if not re.match('^[a-z0-9]+[\w-]+[a-z0-9]+$', option_name,
+            if not re.match(r'^[a-z0-9]+(?:[\w-]*[a-z0-9]+)?$', option_name,
                             flags=re.IGNORECASE):
                 self.err('config.yaml: %s does not conform to naming pattern'
                          % option_name)
@@ -198,13 +214,14 @@ class CharmLinter(Linter):
                         option_name, ', '.join(sorted(missing_keys))))
             invalid_keys = existing_keys - KNOWN_OPTION_KEYS
             if invalid_keys:
-                invalid_keys = [str(key) for key in sorted(invalid_keys)]
+                invalid_keys = sorted([str(key) for key in invalid_keys])
                 self.warn(
                     'config.yaml: option %s has unknown keys: %s' % (
                         option_name, ', '.join(invalid_keys)))
 
             if 'description' in existing_keys:
-                if not isinstance(option_value['description'], basestring) or \
+                if not isinstance(option_value['description'],
+                                  six.string_types) or \
                         option_value['description'].strip() == '':
                     self.warn(
                         'config.yaml: description of option %s should be a '
@@ -235,8 +252,9 @@ class CharmLinter(Linter):
 
 
 class Charm(object):
-    def __init__(self, path):
+    def __init__(self, path, linter=None):
         self.charm_path = path
+        self.linter = linter or CharmLinter()
         if not self.is_charm():
             raise Exception('Not a Charm')
 
@@ -244,7 +262,7 @@ class Charm(object):
         return os.path.isfile(os.path.join(self.charm_path, 'metadata.yaml'))
 
     def proof(self):
-        lint = CharmLinter()
+        lint = self.linter
         charm_name = self.charm_path
         if os.path.isdir(charm_name):
             charm_path = charm_name
@@ -258,41 +276,66 @@ class Charm(object):
 
         hooks_path = os.path.join(charm_path, 'hooks')
         actions_path = os.path.join(charm_path, 'actions')
+        functions_path = os.path.join(charm_path, 'functions')
         yaml_path = os.path.join(charm_path, 'metadata.yaml')
         actions_yaml_file = os.path.join(charm_path, 'actions.yaml')
+        functions_yaml_file = os.path.join(charm_path, 'functions.yaml')
+        layer_yaml_file = os.path.join(charm_path, 'layer.yaml')
         try:
-            yamlfile = open(yaml_path, 'r')
+            yamlfile = open(yaml_path, "rb")
             try:
                 charm = yaml.safe_load(yamlfile)
             except Exception as e:
                 lint.crit('cannot parse ' + yaml_path + ":" + str(e))
                 return lint.lint, lint.exit_code
-
             yamlfile.close()
 
+            proof_extensions = {}
+            try:
+                with open(layer_yaml_file, "rb") as yamlfile:
+                    try:
+                        layer_data = yaml.safe_load(yamlfile)
+                        proof_extensions.update(layer_data.get('proof', {}))
+                    except Exception as e:
+                        lint.warn('cannot parse {}: {}'.format(layer_yaml_file,
+                                                               str(e)))
+            except Exception:
+                # not all charms have a layer.yaml
+                pass
+
             for key in charm.keys():
                 if key not in KNOWN_METADATA_KEYS:
                     lint.err("Unknown root metadata field (%s)" % key)
 
+            for key in REQUIRED_METADATA_KEYS:
+                if key not in charm:
+                    lint.err("Missing required metadata field (%s)" % key)
+
             charm_basename = os.path.basename(charm_path)
-            if charm['name'] != charm_basename:
+            if charm.get('name') != charm_basename:
                 msg = (
                     "metadata name (%s) must match directory name (%s)"
-                    " exactly for local deployment.") % (charm['name'],
+                    " exactly for local deployment.") % (charm.get('name'),
                                                          charm_basename)
                 lint.info(msg)
 
             # summary should be short
-            if len(charm['summary']) > 72:
+            if len(charm.get('summary', '')) > 72:
                 lint.warn('summary should be less than 72')
 
+            validate_display_name(charm, lint)
             validate_maintainer(charm, lint)
             validate_categories_and_tags(charm, lint)
-            validate_storage(charm, lint)
+            validate_storage(charm, lint, proof_extensions.get('storage'))
+            validate_devices(charm, lint, proof_extensions.get('devices'))
             validate_series(charm, lint)
             validate_min_juju_version(charm, lint)
             validate_extra_bindings(charm, lint)
-            validate_payloads(charm, lint)
+            validate_payloads(charm, lint, proof_extensions.get('payloads'))
+            validate_terms(charm, lint)
+            validate_resources(charm, lint, proof_extensions.get('resources'))
+            validate_deployment(charm, lint,
+                                proof_extensions.get('deployment'))
 
             if not os.path.exists(os.path.join(charm_path, 'icon.svg')):
                 lint.info("No icon.svg file.")
@@ -301,9 +344,10 @@ class Charm(object):
                 template_sha1 = hashlib.sha1()
                 icon_sha1 = hashlib.sha1()
                 try:
-                    with open(TEMPLATE_ICON) as ti:
+                    with open(TEMPLATE_ICON, 'rb') as ti:
                         template_sha1.update(ti.read())
-                        with open(os.path.join(charm_path, 'icon.svg')) as ci:
+                        icon_file = os.path.join(charm_path, 'icon.svg')
+                        with open(icon_file, 'rb') as ci:
                             icon_sha1.update(ci.read())
                     if template_sha1.hexdigest() == icon_sha1.hexdigest():
                         lint.info("Includes template icon.svg file.")
@@ -372,7 +416,7 @@ class Charm(object):
                     requires = charm.get('requires')
                     if requires is not None:
                         found_scope_container = False
-                        for rel_name, rel in requires.iteritems():
+                        for rel in six.itervalues(requires):
                             if 'scope' in rel:
                                 if rel['scope'] == 'container':
                                     found_scope_container = True
@@ -414,12 +458,22 @@ class Charm(object):
                 lint.check_hook('config-changed', hooks_path)
 
             if os.path.exists(actions_yaml_file):
-                with open(actions_yaml_file) as f:
+                # TODO: remove action code when it's deprecated.
+                with open(actions_yaml_file, "rb") as f:
                     try:
                         actions = yaml.safe_load(f.read())
+                        validate_actions(actions, actions_path, lint)
                     except Exception as e:
-                        lint.crit('cannot parse ' + actions_yaml_file + ":" + str(e))
-                    validate_actions(actions, actions_path, lint)
+                        lint.crit('cannot parse {}: {}'.format(
+                            actions_yaml_file, e))
+            elif os.path.exists(functions_yaml_file):
+                with open(functions_yaml_file, "rb") as f:
+                    try:
+                        functions = yaml.safe_load(f.read())
+                        validate_functions(functions, functions_path, lint)
+                    except Exception as e:
+                        lint.crit('cannot parse {}: {}'.format(
+                            functions_yaml_file, e))
 
         except IOError:
             lint.err("could not find metadata file for " + charm_name)
@@ -443,7 +497,7 @@ class Charm(object):
 
     def metadata(self):
         metadata = None
-        with open(os.path.join(self.charm_path, 'metadata.yaml')) as f:
+        with open(os.path.join(self.charm_path, 'metadata.yaml'), "rb") as f:
             metadata = yaml.safe_load(f.read())
 
         return metadata
@@ -452,6 +506,11 @@ class Charm(object):
         pass
 
 
+class StrictMapping(colander.Mapping):
+    def __init__(self):
+        super(StrictMapping, self).__init__(unknown='raise')
+
+
 class Boolean(object):
     def deserialize(self, node, cstruct):
         if cstruct is colander.null:
@@ -463,9 +522,94 @@ class Boolean(object):
                 node, '"%s" is not one of true, false' % cstruct)
 
 
+class StringOrEmpty(colander.String):
+    """
+    Subclass of colander.String that defaults allow_empty to True.
+    """
+    def __init__(self, encoding=None, allow_empty=True):
+        super(StringOrEmpty, self).__init__(encoding, allow_empty)
+
+
+class SchemaBuilder(dict2colander.SchemaBuilder):
+    def __init__(self):
+        super(SchemaBuilder, self).__init__()
+        self.add_type('Mapping', StrictMapping)
+        self.add_type('Boolean', Boolean)
+        # override String type to default to allow empty
+        self.add_type('String', StringOrEmpty)
+
+    def _get_validation_schema(self):
+        # override validation schema to ensure that the `missing`
+        # field allows empty strings
+        root = super(SchemaBuilder, self)._get_validation_schema()
+        node_for_missing = [node
+                            for node in root.children
+                            if node.name == 'missing'][0]
+        node_for_missing.typ = StringOrEmpty()
+        return root
+
+
+class ResourceItem(colander.MappingSchema):
+    def schema_type(self, **kw):
+        return StrictMapping()
+
+    type_ = colander.SchemaNode(
+        colander.String(),
+        validator=colander.OneOf(['file', 'oci-image']),
+        name='type',
+    )
+    filename = colander.SchemaNode(
+        colander.String(),
+        missing=colander.drop,
+    )
+    description = colander.SchemaNode(
+        colander.String(allow_empty=True),
+        missing='',
+    )
+
+
+class DevicesItem(colander.MappingSchema):
+    def schema_type(self, **kw):
+        return StrictMapping()
+
+    type_ = colander.SchemaNode(
+        colander.String(),
+        name='type',
+    )
+
+    count = colander.SchemaNode(
+        colander.Integer(),
+        missing=1,
+    )
+
+
+class DeploymentItem(colander.MappingSchema):
+    def schema_type(self, **kw):
+        return StrictMapping()
+
+    type_ = colander.SchemaNode(
+        colander.String(),
+        validator=colander.OneOf(['stateless', 'stateful', 'daemon']),
+        name='type',
+        missing='',
+    )
+
+    service = colander.SchemaNode(
+        colander.String(),
+        validator=colander.OneOf(['loadbalancer', 'cluster', 'omit']),
+        name='service',
+    )
+
+    min_version = colander.SchemaNode(
+        colander.String(),
+        name='min-version',
+        missing='',
+    )
+
+
 class StorageItem(colander.MappingSchema):
     def schema_type(self, **kw):
-        return colander.Mapping(unknown='raise')
+        return StrictMapping()
 
     type_ = colander.SchemaNode(
         colander.String(),
@@ -503,7 +647,7 @@ class StorageItem(colander.MappingSchema):
     @colander.instantiate(missing={})
     class multiple(colander.MappingSchema):
         def schema_type(self, **kw):
-            return colander.Mapping(unknown='raise')
+            return StrictMapping()
 
         range_ = colander.SchemaNode(
             colander.String(),
@@ -518,7 +662,7 @@ class StorageItem(colander.MappingSchema):
 
 class PayloadItem(colander.MappingSchema):
     def schema_type(self, **kw):
-        return colander.Mapping(unknown='raise')
+        return StrictMapping()
 
     type_ = colander.SchemaNode(
         colander.String(),
@@ -527,6 +671,107 @@ class PayloadItem(colander.MappingSchema):
     )
 
 
+def _try_proof_extensions(e, proof_extensions):
+    if not proof_extensions:
+        raise e
+    proof_extensions = {'name': 'extensions',
+                        'type': 'Mapping',
+                        'subnodes': proof_extensions}
+    new_e = colander.Invalid(e.node)
+    for child_e in e.children:
+        if isinstance(child_e, colander.UnsupportedFields):
+            try:
+                schema_builder = SchemaBuilder()
+                schema = schema_builder.dict_to_schema(proof_extensions)
+                schema.deserialize(child_e.fields)
+            except colander.Invalid as ext_e:
+                new_e.add(ext_e)
+        else:
+            new_e.add(child_e)
+    if new_e.children:
+        raise new_e
+
+
+def validate_terms(charm, linter):
+    """Validate terms in charm metadata.
+
+    :param charm: dict of charm metadata parsed from metadata.yaml
+    :param linter: :class:`CharmLinter` object to which info/warning/error
+        messages will be written
+
+    """
+    if 'terms' not in charm:
+        return
+
+    if not isinstance(charm['terms'], list):
+        linter.err('terms: must be a list of term ids')
+
+
+def validate_resources(charm, linter, proof_extensions=None):
+    """Validate resources in charm metadata.
+
+    :param charm: dict of charm metadata parsed from metadata.yaml
+    :param linter: :class:`CharmLinter` object to which info/warning/error
+        messages will be written
+    :param proof_extensions: optional dict of extensions to proof which will
+        only be applied to otherwise unrecognized fields; this allows for
+        layer-defined annotations to pass proof
+
+    """
+    if 'resources' not in charm:
+        return
+
+    if (not isinstance(charm['resources'], dict) or
+            not charm['resources']):
+        linter.err('resources: must be a dictionary of resource definitions')
+        return
+
+    schema = colander.SchemaNode(colander.Mapping())
+    for resource_def in charm['resources']:
+        schema.add(ResourceItem(name=resource_def))
+
+    try:
+        try:
+            schema.deserialize(charm['resources'])
+        except colander.Invalid as e:
+            _try_proof_extensions(e, proof_extensions)
+    except colander.Invalid as e:
+        for k, v in e.asdict().items():
+            linter.err('resources.{}: {}'.format(k, v))
+
+
+def validate_deployment(charm, linter, proof_extensions=None):
+    """Validate deployment in charm metadata.
+
+    :param charm: dict of charm metadata parsed from metadata.yaml
+    :param linter: :class:`CharmLinter` object to which info/warning/error
+        messages will be written
+
+    """
+
+    deployment = charm.get('deployment', {})
+    if deployment == {}:
+        return
+
+    if not isinstance(deployment, dict):
+        linter.err('deployment: must be a dict of config')
+        return
+
+    deployment = dict(deployment=deployment)
+    schema = colander.SchemaNode(colander.Mapping())
+    for item in deployment:
+        schema.add(DeploymentItem(name=item))
+
+    try:
+        try:
+            schema.deserialize(deployment)
+        except colander.Invalid as e:
+            _try_proof_extensions(e, proof_extensions)
+    except colander.Invalid as e:
+        for k, v in e.asdict().items():
+            linter.err('deployment.{}: {}'.format(k, v))
+
+
 def validate_extra_bindings(charm, linter):
     """Validate extra-bindings in charm metadata.
 
@@ -580,18 +825,20 @@ def validate_series(charm, linter):
 
     """
     if 'series' not in charm:
-        return
-
-    if not isinstance(charm['series'], list):
+        linter.err('missing series: must be a list of series names')
+    elif not isinstance(charm['series'], list):
         linter.err('series: must be a list of series names')
 
 
-def validate_storage(charm, linter):
+def validate_storage(charm, linter, proof_extensions=None):
     """Validate storage configuration in charm metadata.
 
     :param charm: dict of charm metadata parsed from metadata.yaml
     :param linter: :class:`CharmLinter` object to which info/warning/error
         messages will be written
+    :param proof_extensions: optional dict of extensions to proof which will
+        only be applied to otherwise unrecognized fields; this allows for
+        layer-defined annotations to pass proof
 
     """
     if 'storage' not in charm:
@@ -607,18 +854,57 @@ def validate_storage(charm, linter):
         schema.add(StorageItem(name=storage_def))
 
     try:
-        schema.deserialize(charm['storage'])
+        try:
+            schema.deserialize(charm['storage'])
+        except colander.Invalid as e:
+            _try_proof_extensions(e, proof_extensions)
     except colander.Invalid as e:
         for k, v in e.asdict().items():
             linter.err('storage.{}: {}'.format(k, v))
 
 
-def validate_payloads(charm, linter):
+def validate_devices(charm, linter, proof_extensions=None):
+    """Validate devices configuration in charm metadata.
+
+    :param charm: dict of charm metadata parsed from metadata.yaml
+    :param linter: :class:`CharmLinter` object to which info/warning/error
+        messages will be written
+    :param proof_extensions: optional dict of extensions to proof which will
+        only be applied to otherwise unrecognized fields; this allows for
+        layer-defined annotations to pass proof
+
+    """
+    devices = charm.get('devices', {})
+    if devices == {}:
+        return
+
+    if not isinstance(devices, dict):
+        linter.err('devices: must be a dictionary of device definitions')
+        return
+
+    schema = colander.SchemaNode(colander.Mapping())
+    for dev in devices:
+        schema.add(DevicesItem(name=dev))
+
+    try:
+        try:
+            schema.deserialize(devices)
+        except colander.Invalid as e:
+            _try_proof_extensions(e, proof_extensions)
+    except colander.Invalid as e:
+        for k, v in e.asdict().items():
+            linter.err('devices.{}: {}'.format(k, v))
+
+
+def validate_payloads(charm, linter, proof_extensions=None):
     """Validate paylaod configuration in charm metadata.
 
     :param charm: dict of charm metadata parsed from metadata.yaml
     :param linter: :class:`CharmLinter` object to which info/warning/error
         messages will be written
+    :param proof_extensions: optional dict of extensions to proof which will
+        only be applied to otherwise unrecognized fields; this allows for
+        layer-defined annotations to pass proof
 
     """
     if 'payloads' not in charm:
@@ -634,7 +920,10 @@ def validate_payloads(charm, linter):
         schema.add(PayloadItem(name=payload_def))
 
     try:
-        schema.deserialize(charm['payloads'])
+        try:
+            schema.deserialize(charm['payloads'])
+        except colander.Invalid as e:
+            _try_proof_extensions(e, proof_extensions)
     except colander.Invalid as e:
         for k, v in e.asdict().items():
             linter.err('payloads.{}: {}'.format(k, v))
@@ -669,6 +958,35 @@ def validate_actions(actions, action_hooks, linter):
             linter.err('actions.{0}: actions/{0} is not executable'.format(k))
 
 
+def validate_functions(functions, function_hooks, linter):
+    """Validate functions in a charm.
+
+    :param functions: dict of charm functions parsed from functions.yaml
+    :param function_hooks: path of charm's /functions/ directory
+    :param linter: :class:`CharmLinter` object to which info/warning/error
+        messages will be written
+
+    """
+
+    if not functions:
+        return
+
+    if not isinstance(functions, dict):
+        linter.err('functions: must be a dictionary of json schemas')
+        return
+
+    # TODO: Schema validation
+    for k in functions:
+        if k.startswith('juju'):
+            linter.err('functions.{}: juju is a reserved namespace'.format(k))
+            continue
+        h = os.path.join(function_hooks, k)
+        if not os.path.isfile(h):
+            linter.warn('functions.{0}: functions/{0} does not exist'.format(k))
+        elif not os.access(h, os.X_OK):
+            linter.err('functions.{0}: functions/{0} is not executable'.format(k))
+
+
 def validate_maintainer(charm, linter):
     """Validate maintainer info in charm metadata.
 
@@ -701,11 +1019,10 @@ def validate_maintainer(charm, linter):
 
     for maintainer in maintainers:
         (name, address) = email.utils.parseaddr(maintainer)
-        formatted = email.utils.formataddr((name, address))
-        if formatted.replace('"', '') != maintainer:
+        if not (name and address):
             linter.warn(
                 'Maintainer format should be "Name <Email>", '
-                'not "%s"' % formatted)
+                'not "%s"' % maintainer)
 
 
 def validate_categories_and_tags(charm, linter):
@@ -731,33 +1048,3 @@ def validate_categories_and_tags(charm, linter):
             'Categories are being deprecated in favor of tags. '
             'Please rename the "categories" field to "tags".'
         )
-
-
-def remote():
-    lp = Launchpad.login_anonymously('charm-tools', 'production',
-                                     version='devel')
-    charm = lp.distributions['charms']
-    current_series = str(charm.current_series).split('/').pop()
-    branches = charm.getBranchTips()
-    charms = []
-
-    for branch in branches:
-        try:
-            branch_series = str(branch[2][0]).split('/')[0]
-            charm_name = str(branch[0]).split('/')[3]
-        except IndexError:
-            branch_series = ''
-        if branch_series == current_series:
-            charms.append("lp:charms/%s" % charm_name)
-        else:
-            charms.append("lp:%s" % branch[0])
-    return charms
-
-
-def local(directory):
-    '''Show charms that actually exist locally. Different than Mr.list'''
-    local_charms = []
-    for charm in os.listdir(directory):
-        if os.path.exists(os.join(directory, charm, '.bzr')):
-            local_charms.append(charm)
-    return local_charms
